import sys
import math
import random
import itertools
import pygame
import data
import const


class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('Brick War')
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        self.game()
    
    def game(self):
        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath('8bp107-01-linde-galileon.ogg'))
            pygame.mixer.music.play(-1)
        except:
            print 'Cannot load sound. :('
        Intro(self, self.screen).loop()


class Intro(object):
    
    def __init__(self, window, screen):
        self.window = window
        self.screen = screen
        self.clock = pygame.time.Clock()
        
        pygame.font.init()
        self.font_18 = pygame.font.Font(data.filepath('square.ttf'), 18)
        self.font_48 = pygame.font.Font(data.filepath('square.ttf'), 48)

    def loop(self):
        self.bg = pygame.Surface((const.WIDTH, const.HEIGHT))
        pygame.draw.rect(self.bg, const.COLOR1,
            pygame.Rect(0,0,const.WIDTH,const.HEIGHT/2))
        pygame.draw.rect(self.bg, const.COLOR2,
            pygame.Rect(0,const.HEIGHT/2,const.WIDTH,const.HEIGHT/2))
        pygame.display.update()
        left_arrow = pygame.image.load(data.filepath('left_arrow.png')).convert()
        right_arrow = pygame.image.load(data.filepath('right_arrow.png')).convert()

        self.hard_mode = False
        
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)

            self.screen.blit(self.bg, (0, 0))
            self.screen.blit(self.font_48.render('Brick War', True, const.COLOR2), (80, 25))
            self.screen.blit(left_arrow, (155, 170))
            self.screen.blit(right_arrow, (155, 400))
            if not self.hard_mode:
                self.screen.blit(self.font_18.render('Press H for harder mode', True, const.COLOR1), (95, 530))
            else:
                self.screen.blit(self.font_18.render('Harder mode activated', True, const.COLOR1), (95, 530))
            self.screen.blit(self.font_48.render('Pick a side', True, const.COLOR1), (85, 580))
            
            dt = self.clock.tick(const.FPS) / 1000.0
                        
            pygame.display.flip()

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            sys.exit()
        elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
            Game(self.window, self.screen, True, self.hard_mode).loop()
        elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
            Game(self.window, self.screen, False, self.hard_mode).loop()
        elif event.key == pygame.K_h:
            self.hard_mode = True

class Game(object):
    
    def __init__(self, window, screen, top, hard_mode):
        self.window = window
        self.screen = screen
        self.top = top
        self.hard_mode = hard_mode
        
        self.clock = pygame.time.Clock()

        self.zones = Zones()

        self.player = Character((320, 320), self.zones)
        self.char_con = pygame.sprite.Group()
        self.char_con.add(self.player)
        self.bullet_mgr = BulletManager(self.player, self.hard_mode)
        self.limits = pygame.sprite.Group()
        self.limits.add(Limit((0,39)))
        self.limits.add(Limit((0,599)))

        self.enemy_mgr = EnemyManager(self.top, self.hard_mode)
        
        self.collision_mgr = CollisionManager(self.player,
            self.bullet_mgr, self.enemy_mgr, self.zones)
        
        pygame.font.init()
        self.font_18 = pygame.font.Font(data.filepath('square.ttf'), 18)
        self.font_48 = pygame.font.Font(data.filepath('square.ttf'), 48)
        
        self.info = Info(self.font_18, self.top)
        self.info_con = pygame.sprite.Group()
        self.info_con.add(self.info)
        
        self.paused = False
        
    def loop(self):
        self.bg = pygame.Surface((const.WIDTH, const.HEIGHT))
        self.bg.fill((100,100,200))
        
        pygame.display.update()
        
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)
            
            if not self.paused:
                self.screen.blit(self.zones.image, (0, 0))
            
            dt = self.clock.tick(const.FPS) / 1000.0
            
            if not self.paused:
                self.player.update(dt)
                self.bullet_mgr.update(dt)
                self.enemy_mgr.update(dt)
                self.collision_mgr.update(dt)
                self.zones.update(dt)
                self.info.update(dt)
            
            if not self.paused:
                self.bullet_mgr.draw(self.screen)
                self.char_con.draw(self.screen)
                self.enemy_mgr.draw(self.screen)
                self.enemy_mgr.arrows.draw(self.screen)
                self.limits.draw(self.screen)
                self.info_con.draw(self.screen)
            
            if self.info.done: self.info_con.remove(self.info)
            
            pygame.display.flip()
            
            if not self.paused and self.zones.done:
                if self.top:
                    if not self.zones.top: self.win()
                    else: self.lose()
                else:
                    if self.zones.top: self.win()
                    else: self.lose()
    
    def pause(self):
        self.paused = not self.paused
    
    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            if self.paused:
                sys.exit()
            else:
                self.pause()
        elif event.key == pygame.K_p:
            self.pause()
        elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
            self.player.move_left()
        elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
            self.player.move_right()
                
    def on_keyup(self, event):
        pass
    
    def win(self):
        self.paused = True
        if not self.top:
            pygame.draw.rect(self.screen, const.COLOR2, pygame.Rect(55, 535, 300, 60))
            self.screen.blit(self.font_48.render('You win! Nice!', True, const.COLOR1), (60, 540))
        else:
            pygame.draw.rect(self.screen, const.COLOR1, pygame.Rect(55, 35, 300, 60))
            self.screen.blit(self.font_48.render('You win! Nice!', True, const.COLOR2), (60, 40))
    
    def lose(self):
        self.paused = True
        if self.top:
            pygame.draw.rect(self.screen, const.COLOR2, pygame.Rect(75, 535, 250, 60))
            self.screen.blit(self.font_48.render('You lose...', True, const.COLOR1), (80, 540))
        else:
            pygame.draw.rect(self.screen, const.COLOR1, pygame.Rect(75, 35, 250, 60))
            self.screen.blit(self.font_48.render('You lose...', True, const.COLOR2), (80, 40))


class Character(pygame.sprite.Sprite):
    
    def __init__(self, pos, zones):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.respawn_pos = list(pos)
        self.image = pygame.image.load(data.filepath('player.png')).convert()
        self.image.set_colorkey((255,255,255))
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.moving_left = False
        self.moving_right = False
        self.vel = [0,0]
        self.xpos = 2
        self.zones = zones

    def start_move_left(self): self.moving_left = True
    def stop_move_left(self): self.moving_left = False
    def start_move_right(self): self.moving_right = True
    def stop_move_right(self): self.moving_right = False
    
    def move_left(self):
        if self.xpos > 0:
            self.xpos -= 1
    
    def move_right(self):
        if self.xpos < 4:
            self.xpos += 1

    def update(self, dt):
        self.pos[0] = 79 * self.xpos + 39
        self.pos[1] = 320 - self.zones.levels[self.xpos]*40
        self.rect.center = self.pos
    
    def update_move(self, dt):
        self.vel = [0,0]
        if self.moving_left:
            self.vel[0] += -self.speed
        if self.moving_right:
            self.vel[0] += self.speed
    

class BulletManager(pygame.sprite.RenderUpdates):
    
    def __init__(self, character, hard_mode):
        pygame.sprite.RenderUpdates.__init__(self)
        self.character = character
        self.shooting_timer = 0
        self.shooting = True
        self.rate_of_fire = 0.6
        if hard_mode: self.rate_of_fire = 0.3
        self.speed = 2000
        self.up_counter = 0
    
    def generate_bullet(self, pos, vel, good):
        bullet = Bullet(pos, vel, good)
        self.add(bullet)
    
    def update(self, dt):
        if self.shooting_timer > 0:
            self.shooting_timer -= dt
        if self.shooting:
            if self.shooting_timer <= 0:
                if self.up_counter in [0,1]:
                    self.generate_bullet(self.character.rect.center, (0,-self.speed), False)
                else:
                    self.generate_bullet(self.character.rect.center, (0,self.speed), True)
                self.shooting_timer = self.rate_of_fire
                self.up_counter = (self.up_counter + 1) % 4
        
        for bullet in iter(self):
            bullet.update(dt)
            if bullet.done:
                self.remove(bullet)


class Bullet(pygame.sprite.Sprite):
    
    def __init__(self, pos, vel, good):
        pygame.sprite.Sprite.__init__(self)
        self.init_pos = list(pos)
        self.pos = list(pos)
        self.vel = list(vel)
        if good:
            self.image = pygame.image.load(data.filepath('punch_bottom.png')).convert()
            self.rect = self.image.get_rect()
            self.rect.midbottom = self.pos
        else:
            self.image = pygame.image.load(data.filepath('punch_top.png')).convert()
            self.rect = self.image.get_rect()
            self.rect.midtop = self.pos
        self.image.set_colorkey((255,255,255))
        self.done = False
        self.good = good
        self.stop = False
        self.stop_timer = 0.8

    def update(self, dt):
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        if ((self.good and self.pos[1] >= self.init_pos[1] + self.image.get_height()) or
            (not self.good and self.pos[1] <= self.init_pos[1] - self.image.get_height())):
            if self.good:
                self.pos[1] = self.init_pos[1] + self.image.get_height()
            else:
                self.pos[1] = self.init_pos[1] - self.image.get_height()
            self.stop = True
            self.vel[1] = 0
        if self.stop:
            self.stop_timer -= dt
            if self.stop_timer <= 0:
                self.done = True
        if self.good:
            self.rect.midbottom = self.pos
        else:
            self.rect.midtop = self.pos
    
    def hit(self):
        self.done = True


class EnemyManager(pygame.sprite.RenderUpdates):
    
    def __init__(self, player_top, hard_mode):
        pygame.sprite.RenderUpdates.__init__(self)
        self.wave_timer_top = 3
        self.wave_timer_bottom = 3
        self.skip = False
        self.player_top = player_top
        self.arrows = pygame.sprite.RenderUpdates()
        self.hard_mode = hard_mode
    
    def generate_enemies(self, positions, top):
        for pos in positions:
            y = 80
            if top:
                if self.player_top: v = 120
                else: v = 60
                enemy = Enemy(pos, (pos * 79, -y), (0, v), top)
            else:
                if not self.player_top: v = 120
                else: v = 60
                enemy = Enemy(pos, (pos * 79, const.HEIGHT + y), (0, -v), top)
            self.add(enemy)
            self.arrows.add(enemy.arrow)
    
    def get_random_pattern(self, top):
        if self.hard_mode:
            quantity = random.randint(2,4)
        else:
            quantity = random.randint(1,3)
        pattern = []
        for i in range(quantity):
            r = random.randint(0,4)
            while r in pattern:
                r = random.randint(0,4)
            pattern.append(r)
        return pattern
    
    def update(self, dt):
        self.wave_timer_top -= dt
        if self.wave_timer_top <= 0:
            if not self.player_top: self.wave_timer_top = 3
            else: self.wave_timer_top = 2
            self.generate_enemies(self.get_random_pattern(True), True)
        self.wave_timer_bottom -= dt
        if self.wave_timer_bottom <= 0:
            if self.player_top: self.wave_timer_bottom = 3
            else: self.wave_timer_bottom = 2
            self.generate_enemies(self.get_random_pattern(False), False)
        for enemy in self:
            enemy.update(dt)
            if enemy.top and enemy.rect.bottom >= 0:
                self.arrows.remove(enemy.arrow)
            if not enemy.top and enemy.rect.top < const.HEIGHT:
                self.arrows.remove(enemy.arrow)
            if enemy.done:
                self.remove(enemy)
                self.arrows.remove(enemy.arrow)


class Enemy(pygame.sprite.Sprite):
    
    def __init__(self, xpos, pos, vel, top):
        pygame.sprite.Sprite.__init__(self)
        
        self.xpos = xpos
        self.pos = list(pos)
        self.vel = vel
        if top:
            self.image = pygame.image.load(data.filepath('enemy_top.png')).convert()
        else:
            self.image = pygame.image.load(data.filepath('enemy_bottom.png')).convert()
        self.rect = self.image.get_rect()
        self.rect.midleft = self.pos
        
        self.top = top
        self.health = 1
        
        self.done = False
        
        if top:
            self.arrow = Arrow((self.xpos * 79 + 39, 20), top)
        else:
            self.arrow = Arrow((self.xpos * 79 + 39, 620), top)
    
    def update(self, dt):
        if self.done: return
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        self.rect.midleft = self.pos
    
    def hit(self):
        self.health -= 1
        if self.health == 0:
            self.die()
    
    def die(self):
        self.done = True


class Arrow(pygame.sprite.Sprite):
    
    def __init__(self, pos, top):
        pygame.sprite.Sprite.__init__(self)
        
        if top:
            self.image = pygame.image.load(data.filepath('arrow_down.png')).convert()
        else:
            self.image = pygame.image.load(data.filepath('arrow_up.png')).convert()
        self.image.set_colorkey((255,255,255))
        self.rect = self.image.get_rect()
        self.rect.center = list(pos)        


class CollisionManager(object):
    
    def __init__(self, character, char_bullet_mgr, enemy_mgr, zones):
        self.character = character
        self.char_bullet_mgr = char_bullet_mgr
        self.enemy_mgr = enemy_mgr
        self.zones = zones
    
    def update(self, dt):
        for bullet in self.char_bullet_mgr:
            collisions = pygame.sprite.spritecollide(bullet,
                self.enemy_mgr, False, pygame.sprite.collide_rect)
            hit = False
            for enemy in collisions:
                if (enemy.top and not bullet.good) or (not enemy.top and bullet.good):
                    enemy.hit()
                    hit = True
            if hit: bullet.hit()
        
        for enemy in self.enemy_mgr:
            if enemy.top:
                if enemy.rect.bottom >= 320 - self.zones.levels[enemy.xpos] * 40:
                    self.zones.levels[enemy.xpos] += 1
                    enemy.die()
            else:
                if enemy.rect.top <= 320 - self.zones.levels[enemy.xpos] * 40:
                    self.zones.levels[enemy.xpos] -= 1
                    enemy.die()


class Zones(pygame.sprite.Sprite):
    
    def __init__(self):
        self.levels = [0,0,0,0,0]
        self.image = pygame.Surface((const.WIDTH, const.HEIGHT))
        self.update_image()
        self.rect = self.image.get_rect()
        self.rect.topleft = [0,0]
        self.done = False
        self.top = False
        
    def update_image(self):
        self.image = pygame.Surface((const.WIDTH, const.HEIGHT))
        i = 0
        for level in self.levels:
            pygame.draw.rect(self.image, const.COLOR1,
                pygame.Rect(i*79,0,79,320 - level * 40))
            pygame.draw.rect(self.image, const.COLOR2,
                pygame.Rect(i*79,320 - level * 40,79,const.HEIGHT - (320-level*40)))
            i += 1
    
    def update(self, dt):
        self.update_image()
        if 8 in self.levels:
            self.done = True
            self.top = True
        if -8 in self.levels:
            self.done = True
            self.top = False


class Limit(pygame.sprite.Sprite):
    
    def __init__(self, pos):
        pygame.sprite.Sprite.__init__(self)
        self.image = pygame.image.load(data.filepath('limit.png')).convert()
        self.rect = self.image.get_rect()
        self.rect.topleft = list(pos)


class Info(pygame.sprite.Sprite):
    
    def __init__(self, font, top):
        pygame.sprite.Sprite.__init__(self)
        self.image = pygame.Surface((const.WIDTH, const.HEIGHT), pygame.SRCALPHA, 32).convert_alpha()
        self.rect = self.image.get_rect()
        self.rect.topleft = (0,0)
        if top:
            self.image.blit(font.render('Avoid this (destroy bricks above)', True, const.COLOR2), (80, 20))
            self.image.blit(font.render('Build down to reach this', True, const.COLOR1), (80, 605))
        else:
            self.image.blit(font.render('Build up to reach this', True, const.COLOR2), (80, 20))
            self.image.blit(font.render('Avoid this (destroy bricks below)', True, const.COLOR1), (80, 605))
        self.image.blit(font.render('Bricks don\'t hurt you', True, const.COLOR2), (70, 100))
        self.image.blit(font.render('but careful: you are your own nemesis', True, const.COLOR2), (30, 120))
        self.timer = 5
        self.done = False

    def update(self, dt):
        self.timer -= dt
        if self.timer <= 0:
            self.done = True